#!/usr/bin/env python3
#
# Execute searches against the storage engine
#
'''
Copyright (c) 2023 B Tasker

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

    Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import os
import sys
import time
import concurrent.futures

sys.path.append(os.path.join(os.path.dirname(__file__), '../lib'))

import config
import storage
from functools import lru_cache

CONCURRENCY = config.getConcurrency()
STATS = {
    "ALL" : {
        "time_ns" : 0,
        "count" : 0,
        "results" : 0
    },
    "DOC" : {
        "time_ns" : 0,
        "count" : 0,
        "results" : 0
    },
    "IMAGE" : {
        "time_ns" : 0,
        "count" : 0,
        "results" : 0
    }
}


def getStats():
    ''' Return a dict of stats derived from the internal counters
    '''
    global STATS
    res = {}
    
    for search_type in ['ALL', 'DOC', 'IMAGE']:
        res[search_type] = {
            "total_time" : STATS[search_type]['time_ns'],
            "count" : STATS[search_type]['count'],
            "results_selected" : STATS[search_type]['results'],
            "mean_results" : 0,
            "mean_search_time_ns" : 0
            }
        
        if res[search_type]["count"] == 0:
            # No searches, return the initialisation values
            continue
        
        # Calculate some means
        res[search_type]["mean_results"] = res[search_type]["results_selected"] / res[search_type]["count"]
        res[search_type]["mean_search_time_ns"] = round(res[search_type]["total_time"] / res[search_type]["count"])
    
    return res

def init(db_path):
    ''' Initialise everything
    
    '''
    storage.DB_PATH = db_path
    storage.loadIndex()


def searchIndexChunk(idx, filters, matchtypes, idx_num, limit, tag_map, count_only):
    ''' Run a search against a portion of the index
    
    Note that utilities/file_location_listing#30 optimised the order in
    which constraints are applied:
    
    - ext constraint should be applied last (because it's expensive)
    - single item constraints should be applied before term matching
    
    # TODO: this function signature's getting a bit long, we should probably collapse
    some of them into a constraints dict or similar
    '''
    results = []
    count = 0
    for e in idx:
        
        # Check whether this entry is part of the chunk we're supposed
        # to check
        if e['i'] != idx_num:
            continue

        # If the result limit has been reached, bail out
        if limit and count >= limit:
            break
        
        # Check that the entry matches constraints, if not 
        # don't waste time checking any match types
        if not _checkConstraints(e, filters):
            continue
        
        # Test each of the match types
        for m in matchtypes:
            
            if m[0] == "k":
                # These are deliberately seperate conditionals: utilities/file_location_listing#32
                if filters['fu']['matchtype'] in ['ANY', 'tag'] :
                    # We're checking tags
                    found = False
                    for tag_id in e['k']:
                        # See whether the tag_id is
                        # in the pregenerated list of
                        # matching tags
                        if (tag_id in tag_map['ids']
                            and checkIndexConstraint("ext", e, filters['fu'])
                            ):
                            # We've got a match
                            if count_only:
                                count += 1
                                break
                                                        
                            info = storage.getStoredItem(False, e['h'])
                            if not info:
                                # Something went wrong and we didn't get the item
                                print(f"{time.time()} Thread {idx_num} Unable to retrieve item for {e['u']}")
                                break
                            
                            # Check whether we matched against an actual, or implied tag
                            tag_implied = (tag_id in tag_map['implied'])
                            info['tagtype'] = "actual"
                            if tag_implied:
                                info['tagtype'] = 'implied'
                                info['implied-tag'] = tag_map["words"][tag_id]
                            
                            info['matchtype'] = m[1]
                            found = True
                            if not limit or count <= limit:
                                results.append(info)
                                count += 1
                                
                            # Either way, break out - no point checking other tag ids once we found a match
                            break
                    
                    # If we found a match, break out of checking matchtypes
                    if found:
                        break
                # Either way, carry on - we don't want the block below trying to work with keywords
                continue
            
            # don't be case-sensitive
            url = e[m[0]].lower()
            if (filters['fu']['matchtype'] in ['ANY', m[1]] 
                    and checkTerms(url, filters)
                    and checkIndexConstraint("ext", e, filters['fu'])
                    ):
                    # Found a match
                    if count_only:
                        count += 1
                        break
                    
                    # Get info for the match
                    info = storage.getStoredItem(False, e['h'])

                    if not info:
                        # Something went wrong and we didn't get the item
                        print(f"{time.time()} Thread {idx_num} Unable to retrieve item for {e['u']}")
                        break
                    
                    info['matchtype'] = m[1]
                    
                    # Not used for this kind of match
                    info['tagtype'] = 'na'
                    
                    if not limit or count <= limit:
                        results.append(info)
                        count += 1
                    
                    # This deliberately isn't indented into the 
                    # conditional above.
                    #
                    # If we got this far, then most criteria matched
                    # either we entered the conditional above and have
                    # added a result or we failed to meet criteria
                    # If the latter, there's no point checking name/title
                    # because we'll fail the same criteria.
                    break

    if count_only:
        results = [count]

    return results


def _checkConstraints(e, filters):
    ''' Check the index entry against a variety of constraints
    '''
    return (checkContentType(e, filters['fu'])
            and checkIndexConstraint("domain", e, filters['fu'])
            and checkIndexConstraint("hastitle", e, filters['fu'])
            and checkIndexConstraint("prefix", e, filters['fu'])
            )
    

def search(srchterm, search_type="ALL", limit=False, count_only=False):
    ''' Run a search
    
    search_type should be one of ALL, IMAGE or DOC
    '''
    global CONCURRENCY
    global STATS
    filters = processSearchTerm(srchterm)
    
    # Iterate through known tags testing whether any
    # match the provided terms
    
    tag_map = {
        "ids" : [],
        "words" : {},
        "implied" : storage.INDEXES['IMPLIED']
    }
    
    for keyword in storage.INDEXES['TAGS']:
        if checkTerms(keyword, filters):
            i = storage.INDEXES['TAGS'][keyword]
            # Tag matched, push it's ID into tag_ids
            tag_map['ids'].append(i)
            # Map the ID back to the word(s)
            tag_map["words"][i] = keyword
    
    # Define the match-types to try
    matchtypes = [
        ['u', 'url'],
        ['n', 'title']
        ]
    
    if config.getTagsEnabled():
        matchtypes.append(['k', 'tag'])
    
    # Initialise
    results = []
    idx_scan = 0
    storage_load = 0
    start_time = time.time_ns()
    
    # Spin the searches out to threads 
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = {}
        results = []
        for idx_num in range(CONCURRENCY):
            futures[idx_num] = executor.submit(
                searchIndexChunk, 
                storage.INDEXES[search_type], 
                filters, 
                matchtypes,
                idx_num,
                limit,
                tag_map,
                count_only
                )
        
        # Await the results
        for idx_num in range(CONCURRENCY):
            res = futures[idx_num].result()
            #print(f"Thread {idx_num} returned {len(res)} results")
            results = results + res
    
    rescount = len(results)
    
    if count_only:
        results = {"result_count": sum(results), "term": srchterm}
    else:
        # Sort the results before we enforce limits
        results.sort(reverse=True, key=lambda x: x.get("last-mod-s", 0))
        
        if limit and rescount > limit:
            results = results[0:limit]
    
    end_time = time.time_ns()
    runtime = end_time - start_time
    print((f"Processed search\n"
          f"\tFilters: {filters}\n"
          f"\tSearch Type: {search_type}\n"
          f"\tSearch Time: {runtime / 1000000000}s"))
    
    STATS[search_type]["count"] += 1
    STATS[search_type]["time_ns"] += runtime
    STATS[search_type]["results"] += rescount
    
    return results


def checkTerms(term, filters):
    ''' Run either and AND or an OR search
    against the provided term
    '''

    if "mode" in filters['fu'] and filters['fu']['mode'] == "OR":
        r = checkOr(term, filters['OR'])
    else:
        r = checkAnd(term, filters['AND'])
    
    # Apply NOT filters
    return (r and checkNot(term, filters['NOT']))


def checkAnd(term, ands):
    ''' Check whether all AND filters apply to an item
    '''
    
    # if no ands have been provided, nothing can match
    if len(ands) < 1:
        return False
    
    for filt in ands:
        if filt not in term:
            # Term is missing
            return False
    
    # if we got this far, everything matched
    return True


def checkNot(term, nots):
    ''' Cycle through any provided NOT filters and exclude any
    result that contains those substrings
    '''
    
    for filt in nots:
        if filt in term:
            # Filtered term is present
            return False
        
    # if we got this far, nothing matched
    return True


def checkOr(term, ors):
    ''' Check whether any OR filters apply to an item
    '''
        
    for filt in ors:
        if filt in term:
            # Term is present
            return True
    
    # if we got this far, nothing matched
    return False   


def checkContentType(idxitem, fu):
    ''' Check whether we're filtering for a specific content-type
    
    This can be done with index info alone
    '''
    if "content-type" not in fu:
        # No filter applied
        return True
    
    # Otherwise, there's a filter, see if it matches
    return (idxitem['t'] == fu['content-type'])


def checkIndexConstraint(field, idxitem, fu):
    ''' Check whether the index information for the item meets provided
    constraints
    '''
    if field not in fu:
        # No filter applied
        return True

    # hastitle is handled a little differently because its a logic check
    # rather than a simple matching operation
    if field == "hastitle":
        titlen = len(idxitem['n'])
        if titlen > 0 and fu['hastitle'] in ["true","1","y"]:
            return True
        elif titlen == 0 and fu['hastitle'] in ["false","0","n"]:
            return True
        else:
            return False    
    
    # Filename extension matching is also handled a little differently
    if field == "ext":
        p_sp = idxitem["u"].split(".")
        ext = p_sp[-1].lower()
        return (ext == fu["ext"].lstrip("."))
    
    if field == "prefix":
        # Strip the leading slash - the user might or might not
        # have omitted it
        pref = fu['prefix'].lstrip("/")
        return (idxitem['p'].lstrip("/").startswith(pref))
    
    if field == "domain":
        return (idxitem["d"] == fu[field])
    

def checkItemConstraint(field, item, fu):
    ''' Check whether the stored item's recorded field matches the 
    operator provided in search
    '''
    if field not in fu:
        # No filter applied
        return True
    
    # Otherwise, there's a filter, see if it matches
    return (item[field].lower() == fu[field])


# Although it might seem odd to have a LRU cache on this, it does make a
# little sense. The search portal, by default, will run a search twice - once
# for a docs search and again to search for related images
#
# The two sets of terms will be exactly the same, so this saves cutting the
# term up twice.
#
# And, of course, if the search is switched from capped to uncapped, we'll 
# receive the same term *again*.
@lru_cache(maxsize=100)
def processSearchTerm(term):
    '''
        Parse a search term and identify filters that need to be applied
    '''
    mode = 'AND'
    filtered_terms = []
    
    filters = {
        "AND" : [],
        "OR" : [],
        "NOT" : [],
        "fu" : {
            'matchtype' : 'ANY'
            }
        }
    
    terms = term.split(" ")
    dorks = [
        'content-type', 
        'domain', 
        'ext', 
        'hastitle', 
        'matchtype', 
        'mode',
        'prefix'
        ]
    
    for t in terms:
        t = t.lower()
        if t.startswith("-"):
            nt = t.lstrip("-")
            if len(nt) > 0:
                filters["NOT"].append(nt)
                continue

        # It's a search-fu
        t_sp = t.split(":")
        if t_sp[0] in dorks:
            filters["fu"][t_sp[0]] = t_sp[1].strip()
            
            # domain filters are a little special. We don't have the domain
            # in the index except as part of the URL - in order to check the domain
            # exactly we either need to parse the key or fetch the stored item (which
            # does have the domain).
            #
            # Both of those are *really* inefficient
            #
            # Instead, we inject the provided domain as an AND match so that candidates
            # effectively pass through an initial sift.
            #
            # The fu filter will then have fewer items to pull from storage in order to
            # perform exact domain matching
            if t_sp[0] in ["domain", "prefix"]:
                filters["AND"].append(t_sp[1].strip())
            
            if t_sp[0] == "mode":
                if t_sp[1] == "exact":
                    mode = "exact"
                elif t_sp[1] == "or":
                    mode = "OR"

            continue
            
        # Otherwise it's an AND
        if len(t) > 0:
            #filters["AND"].append(t)
            filtered_terms.append(t)
        

    # check whether we're still in AND mode, or if we've flipped over to exact or OR
    if mode == "exact":
        filters["AND"] = [" ".join(filtered_terms)]
    else:
        filters[mode] = filtered_terms
    
    # Set into fu - mainly for logging purposes
    filters["fu"]["mode"] = mode
    
    return filters
