#!/usr/bin/env python3
#
# Storage module
#
'''
Copyright (c) 2023 B Tasker

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

    Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
import hashlib
import gzip
import json
import os
import sys
import tempfile
import time

from functools import lru_cache
from pathlib import Path
from subprocess import Popen
from urllib.parse import urlparse

sys.path.append(os.path.join(os.path.dirname(__file__), '../lib'))

import config
import tombstone


INDEXES = False # This will be populated on first index load
DB_PATH = '/opt/search_db'
INDEX_ALL = False
LOADED_AT = False
CONCURRENCY = config.getConcurrency()

STATS = {
    "store_file_loads" : 0,
    "store_file_fails" : 0,
    "store_load_time_ns" : 0,
    "index_loads" : 0,
    "index_load_time_ns" : 0,
    "index_items_loaded" : 0,
    "index_checksum_mismatches": 0,
    "tags_checksum_mismatches": 0,
    }

def loadIndex():
    ''' Load the index from disk
    
    '''
    
    global LOADED_AT
    global INDEXES
    global STATS
    
    # Do we already have an index loaded in memory?
    if not INDEXES or len(INDEXES["IMAGE"]) > 0 or len(INDEXES["ALL"]) > 0 or len(INDEXES["DOC"]) > 0:
        # Blank it
        INDEXES = {
            "ALL" : [],
            "IMAGE" : [],
            "DOC" : [],
            "TAGS" : {},
            "IMPLIED" : [],
            "tombcount" : 0
        }
    start = time.time_ns()
    
    # Load details of any implied tags
    implied_tags = config.getImpliedTags()
    
    # read the tag index first
    url_tags = {}
    tag_id = 0
    tag_header = False
    if config.getTagsEnabled() and os.path.exists(f"{DB_PATH}/tags"):
        with gzip.open(f"{DB_PATH}/tags", 'r') as f:
            lcount = 0
            # Read the index into a list - we're going to be
            # hashing this later
            tag_list = []
            
            # Iterate through
            for l in f.read().splitlines():
                l_dec = l.decode()
                if lcount == 0:
                    lcount += 1
                    if l_dec[0] == chr(255):
                        # It's a header line
                        tag_header = l_dec
                        continue                
                
                tag_list.append(l_dec)
                # Split the line up
                l_sp = l_dec.split("!#!")
                
                # Extract the tag
                k = l_sp[0]
                
                # Put the end of the line backtogether
                # even in the unlikely event our seperator
                # exists in it
                try:
                    v = json.loads("!#!".join(l_sp[1:]))
                except Exception as e:
                    print(f"Unable to load keys for tag {k}")
                    continue
                
                # Add the tag and it's calculated ID to the
                # tag list
                # check if it's there to prevent dupes
                if k not in INDEXES['TAGS']:
                    INDEXES['TAGS'][k] = tag_id
                    mytag = tag_id
                else:
                    mytag = INDEXES['TAGS'][k]
                
                # Update each of the keys to note it has this tag
                for url in v:
                    if url not in url_tags:
                        url_tags[url] = []
                
                    url_tags[url].append(tag_id)
                
                # Check whether this tag leads to any implied tags
                if k in implied_tags:
                    for implied_tag in implied_tags[k]:
                        # increment the tag id
                        tag_id += 1
                        
                        # insert the tag 
                        INDEXES['TAGS'][implied_tag] = tag_id
                        
                        # Mark the ID as being for an implied tag
                        INDEXES["IMPLIED"].append(tag_id)
                        
                        # mark each fo the urls as being linked to this tag
                        _ = [url_tags[url].append(tag_id) for url in v]
                            
                
                # Increment the tag id
                tag_id += 1
    
    # Check whether the index matches it's header
    if tag_header:
        header_dec = getIndexHeaderItems(tag_header)
        disk_hash = hashEntries(tag_list, skip_first = False)
        # Check the hashes match
        if header_dec['hash'] != disk_hash:
            # We can't do **much** more than warn about it
            # 
            # The alternative is to refuse to come up, which'd be a service outage
            print(f"WARN: Tag checksum mismatch. Stored: {header_dec['hash']} Loaded:{disk_hash}")
            STATS["tags_checksum_mismatches"] += 1

    # Free up memory, we no longer need the list of lines
    del(tag_list)
    
    tombstones = {}
    for tpath in tombstone.getPaths(f"{DB_PATH}/index"):
        if os.path.exists(tpath):
            with open(tpath, 'r') as f:
                for l in f.read().splitlines():
                    tombstones[l.strip()] = 1
                    INDEXES["tombcount"] += 1
    
    seen = {}
    dupes = 0
    lcount = 0
    idx_header = False
    with gzip.open(f"{DB_PATH}/index", 'r') as f:        
          # We mark each entry with an index number
          # this allows the search to parallelise by having
          # threads only consider entries with their corresponding
          # index number
          idx_num = 0
          
          idx_list = []
          
          # Iterate through the loaded file
          for l in f.read().splitlines():
            l_dec = l.decode()
            if lcount == 0:
                lcount += 1
                if l_dec[0] == chr(255):
                    # It's a header line
                    idx_header = l_dec
                    continue
                
            idx_list.append(l_dec)
            index_l = l_dec.split(",!,")
            
            # Has the URL been tombstoned?
            if index_l[0] in tombstones:
                continue
            
            if index_l[0] in seen:
                dupes += 1
                print(f"Skipping duplicate key {index_l[0]}")
                continue

            
            # Extract the file path from the key
            linkparsed = urlparse(index_l[0])
            dom = linkparsed.netloc.lower()
            pth = linkparsed.path.lower()
            
            # Create the index item
            idxitem = {
                "u" : index_l[0], # url
                "h" : index_l[1], # index hash
                "t" : index_l[2], # content-type
                "n" : index_l[3], # title/name
                "i" : idx_num, # Index number
                "p" : pth, # path
                "d" : dom, # domain
                "k" : []
            }
            
            # See whether the key exists in url_tags
            # if so, we've some tag keywords to record against this
            # entry
            if idxitem['u'] in url_tags:
                idxitem['k'] = url_tags[idxitem['u']]
            
            # Append into the relevant index(es)
            
            # Should we be populating the "ALL" index
            #
            # This is off by default - generally speaking we
            # tend to want to search by type, so having ALL is
            # just wasting RAM
            if INDEX_ALL:
                INDEXES["ALL"].append(idxitem)
                
            # Put into either image or doc
            #
            # TODO: do we perhaps want a video index?
            if index_l[2] in ["image/jpeg", "image/png", "image/gif"]:
                INDEXES["IMAGE"].append(idxitem)
            else:
                INDEXES["DOC"].append(idxitem)
             
            # Mark the key as seen
            seen[index_l[0]] = 1
            
            idx_num += 1
            if idx_num >= CONCURRENCY:
                idx_num = 0


    # Check whether the index matches it's header
    if idx_header:
        header_dec = getIndexHeaderItems(idx_header)
        disk_hash = hashEntries(idx_list, skip_first = False)
        # Check the hashes match
        if header_dec['hash'] != disk_hash:
            # We can't do **much** more than warn about it
            # 
            # The alternative is to refuse to come up, which'd be a service outage
            print(f"WARN: Index checksum mismatch. Stored: {header_dec['hash']} Loaded:{disk_hash}")
            STATS["index_checksum_mismatches"] += 1

    # Free up memory, we no longer need the list of lines
    del(idx_list)


    LOADED_AT = time.time()
    INDEXES["last_mod"] = getIndexLastMod()
    stop = time.time_ns()
    loaded_items = len(seen.keys())
    
    # Increment the stats counters
    STATS["index_load_time_ns"] += stop-start
    STATS["index_loads"] += 1
    STATS["index_items_loaded"] += loaded_items
    
    print(f"Index loaded {loaded_items} keys in {stop-start}ns")
    if dupes > 0:
        print(f"WARN: {dupes} duplicate index items detected")
    
    # Flush the file-cache
    readStoredFile.cache_clear()


def deleteStoredItem(url):
    '''
    Return stored item for the url
    '''
    
    # calculate a hash of the URL
    fname = getUrlHash(url)
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    
    file_path = f"{dir_path}/{fname}"
    
    if not os.path.exists(file_path):
        return False
    
    #print(f"Will delete{file_path}")
    os.remove(file_path)
    
    # Check if that's resulted in an empty directory
    if len(os.listdir(dir_path)) == 0:
        try:
            os.rmdir(dir_path)
        except:
            # oh well, no harm done
            pass

    # Also need to check whether there's a thumbnail
    thumbs_path = getThumbsPath(fname)
    path_prefix = f"{thumbs_path}/{fname}"
    for ext in [".gif", ".jpg"]:
        file_path = f"{path_prefix}{ext}"
        if not os.path.exists(file_path):
            continue
        
        # Otherwise, remove it
        os.remove(file_path)
    
    # Tombstone the URL so it'll be ignored in the index
    # at next index reload
    tombstone.append(f"{DB_PATH}/index", url)

        
def getIndexFileStats():
    ''' Run a stat on the index files
    '''
    stats = {
        "index" : {},
        "index.tomb" : {},
        "tags": {}
        }
    
    for f in stats:    
        if os.path.exists(f"{DB_PATH}/{f}"):
            stat = os.stat(f"{DB_PATH}/{f}")
            stats[f]['size'] = stat.st_size
            stats[f]['mtime'] = round(stat.st_mtime)
        
    return stats
    
def getIndexLastMod():
    ''' Get the mtime of the index
    If there's a tombstone **and** it's more recent, that
    time is returned instead
    '''
    
    tomb = 0
    tags = 0
    mtimes = []
    
    for idx_type in ['index', 'tags']:
        if os.path.exists(f"{DB_PATH}/{idx_type}"):
            mtimes.append(os.path.getmtime(f"{DB_PATH}/{idx_type}"))
            
        for tpath in tombstone.getPaths(f"{DB_PATH}/{idx_type}"):
            if os.path.exists(tpath):   
                mtimes.append(os.path.getmtime(tpath))

    return max(mtimes)
    

def getIsIndexFresh():
    ''' Check whether the index has changed since it was loaded
    
    This is achieved by getting mtime and comparing to the mtime
    stored with the index
    '''
    return (getIndexLastMod() == INDEXES['last_mod'])


def getStats():
    ''' Return stats about the storage layer
    '''
    global LOADED_AT
    global STATS
    
    ci = readStoredFile.cache_info()
    
    return {
        "DOC_index_cnt": len(INDEXES["DOC"]),
        "IMAGE_index_cnt": len(INDEXES["IMAGE"]),
        "TAGS_index_cnt" : len(INDEXES["TAGS"]),
        "IMPLIED_index_cnt" : len(INDEXES["IMPLIED"]),
        "TOMB_cnt" : INDEXES['tombcount'],
        "index_last_load" : round(LOADED_AT),
        "load_age_s" : round(time.time() - LOADED_AT),
        "index_last_mod": round(getIndexLastMod()),
        "files" : getIndexFileStats(),
        "warnings" : {
            "index_checksum_mismatches": STATS["index_checksum_mismatches"],
            "tags_checksum_mismatches": STATS["tags_checksum_mismatches"]
            },
        "file_cache" : {
            "size" : ci.currsize,
            "hits" : ci.hits,
            "misses" : ci.misses
            },
        "store_files" : {
            "loads" : STATS["store_file_loads"],
            "fails" : STATS["store_file_fails"],
            "time_ns" : STATS["store_load_time_ns"]
            },
        "index" : {
            "loads" : STATS["index_loads"],
            "time_ns" : STATS["index_load_time_ns"],
            "loaded_entries" : STATS["index_items_loaded"]
            },
        
        }


def getStoredInfo(url):
    # Return stored info for the url
    # TODO
    
    # calculate a hash of the URL
    fname = getUrlHash(url)
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    file_path = f"{dir_path}/{fname}"
    crawler_ver = '0.0.0'
    
    j = readStoredFile(file_path)
    if not j:
        return 'FFFFFF', 'FFFFFF', 'FFFFFF', crawler_ver
    
    if "engine" in j and "version" in j['engine']:
        crawler_ver = j['engine']['version']
    
    # return
    # last_mod, content_hash, valid_date
    # Support for content hashing isn't implemented yet
    return j['last-mod'], 'FFFFFFFFF', j['valid-at'], crawler_ver


def getStoredItem(url, fname=False):
    '''
    Return stored item for the url
    
    If a fname is provided, that's used instead
    of calculating from the URL
    '''
    
    if not fname:
        # calculate a hash of the URL
        fname = getUrlHash(url)
    
    # Calculate dir path
    dir_path = getStoragePath(fname)
    file_path = f"{dir_path}/{fname}"

    # read the file and return the JSON
    return readStoredFile(file_path)


@lru_cache(maxsize=10000)
def readStoredFile(file_path):
    ''' Open a storage file for reading and return the json
    
    Attempts to read the file using gzip (the new storage format)
    and falls back to plain (the old) if that fails
    
    The fallback will eventually be removed.
    '''
    
    if not os.path.exists(file_path):
        return False
    
    global STATS
    start = time.time_ns()
    
    with gzip.open(file_path) as f:
        for line in f:
            pass
        
        # line is now the last line
        try:
            j = json.loads(line)
        except:
            print("Failed to load file")
            STATS["store_file_fails"] += 1
            invalidateFile(file_path)
            return False    
    
    # Update stats
    end = time.time_ns()
    STATS["store_load_time_ns"] += end-start
    STATS["store_file_loads"] += 1
    
    # Return the loaded data
    return j   


def getStoragePath(fname, full=True):
    ''' Calculate the directory path
    
    '''
    parent_dir1 = fname[0:2]
    parent_dir2 = fname[2:4]
    if full:
        return f"{DB_PATH}/store/{parent_dir1}/{parent_dir2}"
    else:
        return f"{parent_dir1}/{parent_dir2}"


def getThumbsPath(fname, full=True):
    ''' Calculate the directory path for a thumbnail
    
    '''
    parent_dir1 = fname[0:2]
    parent_dir2 = fname[2:4]
    if full:
        return f"{DB_PATH}/thumbs/{parent_dir1}/{parent_dir2}"
    else:
        return f"{parent_dir1}/{parent_dir2}"   
    
@lru_cache(maxsize=10000)
def getUrlHash(url):
    ''' Generate a hash of an URL 
    This is used to navigate the cache on disk
    '''
    h = hashlib.sha256()
    h.update(url.encode('utf-8'))
    return h.hexdigest()

def invalidateFile(filepath):
    ''' Mark a file as being invalid
    '''
    print(f"Marking as invalid: {filepath}")
    try:
        os.replace(filepath, f"{filepath}.invalid")    
    except Exception as e:
        print(f"Unable to mark invalid: {e}")
    

def indexStorage():
    ''' After a run, we need to rebuild the index
    
    This is essentially just an ordered list of URLs
    '''
    index_list = []
    keyword_d = {}
    do_tags = config.getTagsEnabled()
    ignored_tags = config.getIgnoredTags()
    
    # Lock the index tombstone
    tombstone.lock(f"{DB_PATH}/index")
    
    for subdir, dirs, files in os.walk(f"{DB_PATH}/store"):
        for file in files:
            # Skip temporary and invalid files
            if file.endswith(".invalid") or file.endswith(".tmp"):
                continue
            
            filepath = os.path.join(subdir, file)
            entry = [False, False, False, False]
            keywords = False
            # Open the file and extract the key and hash
            try:
                with gzip.open(filepath, 'r') as f:
                    last_mod = 0
                    for line in f:
                        line = line.decode()
                        if line.startswith("KEY"):
                            entry[0] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("HASH"):
                            entry[1] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("LAST-MOD"):
                            last_mod = int("".join(line.split(":")[1:]).strip())
                        if line.startswith("CONTENTTYPE"):
                            entry[2] = ":".join(line.split(":")[1:]).strip()
                        if line.startswith("TITLE"):
                            entry[3] = ":".join(line.split(":")[1:]).strip()
                        if do_tags and line.startswith("TAGS"):
                            keywords = ":".join(line.split(":")[1:]).strip()

                            
            except:
                invalidateFile(filepath)
                            
            if entry[2] is False:
                entry[2] = "!unk!"
                
            if entry[3] is False:
                entry[3] = "!unk!"
                
            if keywords:
                for k in json.loads(keywords):
                    k_l = k.lower()
                    if k_l in ignored_tags:
                        # Don't write this tag to the index
                        continue
                    if k_l not in keyword_d:
                        keyword_d[k_l] = []
                    keyword_d[k_l].append(entry[0])
                    
            # We should now have an entry with at least 2 parts. If we don't
            # skip this file
            if len(entry) < 4 or False in entry:
                continue
            
            # The index entry is:
            # url, path
            # Calculate the path
            pth = getStoragePath(entry[1], False)
            index_list.append({"e": f"{entry[0]},!,{entry[1]},!,{entry[2]},!,{entry[3]}", "l": last_mod})
            
    # Sort the list
    index_list.sort(reverse=True, key=lambda x: x.get("l"))
    
    index_out_list = [buildIndexHeader(index_list, 'index')]
    for d in index_list:
        index_out_list.append(d["e"])
    
    # Generate a suffix to append to temporary filenames
    temp_suffix = time.time_ns()

    # Write to the index
    with gzip.open(f"{DB_PATH}/index.{temp_suffix}.tmp", 'wt') as f:
        f.write('\n'.join(index_out_list))
    
    if do_tags:
        tag_list = []
        for tag in keyword_d:
            tag_list.append(f"{tag}!#!{json.dumps(keyword_d[tag])}")

        tag_list.sort()
        tag_list.insert(0, buildIndexHeader(tag_list, 'tags'))

        with gzip.open(f"{DB_PATH}/tags.{temp_suffix}.tmp", 'wt') as f:
            f.write('\n'.join(tag_list))
        
    # Put the indexes into use, but only if they've changed
    for index_type in ['index', 'tags']:
        if getIndexHash(f"{DB_PATH}/{index_type}.{temp_suffix}.tmp") != getIndexHash(f"{DB_PATH}/{index_type}"):
            os.replace(f"{DB_PATH}/{index_type}.{temp_suffix}.tmp", f"{DB_PATH}/{index_type}")
        else:
            print(f"index file {index_type} unchanged")
            os.remove(f"{DB_PATH}/{index_type}.{temp_suffix}.tmp")
            
    # Remove the tombstone file, if it exists
    tombstone.remove(f"{DB_PATH}/index")
    # Unlock
    tombstone.unlock(f"{DB_PATH}/index")


def buildIndexHeader(l, idx_type):
    ''' Build the header line for an index file
    
        Byte 1: index type
    '''
    
    header_entries = [chr(255)]
    
    # Set a byte to indicate index type
    if idx_type == "index":
        header_entries.append(chr(0))
    elif idx_type == "tags":
        header_entries.append(chr(1))
        
    # Set the index version (we don't currently use this
    # but seems prudent to provide the capability in case 
    # it's needed in the future)
    header_entries.append(chr(1))

    # Generate a hash of the entries
    # it's a hex encoded sha1 so will be 40 bytes
    h = header_entries.append(hashEntries(l, False))
    
    return ''.join(header_entries)

    
def hashEntries(l, skip_first = True):
    ''' Iterate through index entries and hash them    
    
        Skip the first line (the header) by default
    '''
    
    # We'll skip a line when lnum = 0
    # if we're not supposed to skip, make sure that's
    # never true
    lnum = int(skip_first == False)
    
    # Iterate through and hash
    hasher = hashlib.sha1()

    for line in l:
        # Skip the first line
        if lnum == 0:
            lnum += 1
            continue
        if isinstance(line, dict):
            line = line['e']
        hasher.update(line.encode('utf-8'))
    
    '''
    a hex digest is slightly less efficient for storage but 
    removes complexity for reading - there's no possibility of
    a newline appearing in it (meaning that we don't have to
    diverge from using Python's build in split_lines())
    '''
    return hasher.hexdigest()


def getIndexHash(fname):
    ''' Load an index file and extract it's hash from the header
    '''
    if not os.path.exists(fname):
        return ''

    with gzip.open(fname, 'rb') as f:
        for line in f.read().splitlines():
            # the hash starts at the 4th byte of the header
            # and is 40 bytes long
            try:
                return getIndexHeaderItems(line.decode())['hash']
            except:
                # We tried to read beyond the end of the line
                # this likely means that we've tried to read the
                # header of an index generated before headers
                # were introduced
                return ''


def getIndexHeaderItems(l):
    ''' Split an index header line up into it's constituents
    '''
    res = {
        "type" : "unknown",
        "version" : 0,
        "hash" : ""
        }
    
    if l[1] == chr(0):
        res["type"] = "index"
    elif l[1] == chr(1):
        res["type"] = "tags"
        
    res["version"] = int(ord(l[2]))
    res["hash"] = l[3:44]
    return res


def storeInfo(pd):
    # Write out to storage 
    # TODO
    # For now, this just prints
    
    # Check what version string should be going into
    # metadata
    storefile_version = config.getStoreFileVersion()
    
    # Serialise the keywords
    kw_line = ""
    if len(pd['keywords']) > 0:
        kw = json.dumps(pd['keywords'])
        kw_line = f"TAGS:{kw}\n"
    
    # Convert the last mod date to an integer
    try:
        ts = time.strptime(pd['last-mod'],'%a, %d %b %Y %H:%M:%S %Z')
        last_mod = int(time.strftime('%s', ts))
    except:
        # Not everything has a last modified date - a lot of auto-generated
        # directory listings don't
        last_mod = 0
    
    file_str = (
        f"KEY:{pd['key']}\n"
        f"HASH:{pd['hashkey']}\n"
        f"DOMAIN:{pd['domain']}\n"
        f"PATH:{pd['path']}\n"
        f"FILE:{pd['filename']}\n"
        f"CONTENTTYPE:{pd['content-type']}\n"
        f"TITLE:{pd['title']}\n"
        f"ENGINE-VER:{storefile_version}\n"
        f"LAST-MOD:{last_mod}\n"
        f"{kw_line}"
        "\n"
    )
    
    pd['engine'] = {
        "version" : storefile_version
        }
    
    # Store the epoch last mod to allow easy sorting
    # at search time
    pd['last-mod-s'] = last_mod
    
    file_str += json.dumps(pd)
    
    #print("===== ENTRY =====")
    #print(file_str)
    #print("===== ENTRY ENDS =====")
    
    # Store to disk
    dir_path = getStoragePath(pd['hashkey'])
    
    # ensure it exists
    path = Path(dir_path)
    path.mkdir(parents=True, exist_ok=True)
    
    # Open the file for writing
    fh = gzip.open(f"{dir_path}/{pd['hashkey']}.tmp", 'wt')
    fh.write(file_str)
    fh.close()
    
    # Move into place
    os.replace(f"{dir_path}/{pd['hashkey']}.tmp", f"{dir_path}/{pd['hashkey']}")
    
    
def storeImage(url, pg, contenttype):
    ''' We've fetched an image
    
    Thumbnail it
    '''
    fname = getUrlHash(url)
    
    thumbs_path = getThumbsPath(fname, full=True)
    
    # ensure it exists
    path = Path(thumbs_path)
    path.mkdir(parents=True, exist_ok=True)    
    
    # Save the main image into a temporary file
    tf = tempfile.NamedTemporaryFile(delete=True)
    tf.write(pg.content)
    tf.seek(0)
    
    if contenttype in ["image/gif"]:
        print(f"Thumbnailing gif {url}")
        # Can't thumnail a gif into a jpg, you just get lots of individual frames
        # Thumbnail it
        out, err = Popen(["convert",tf.name,"-loop", "0", "-resize","x450","-quality","75",f"{thumbs_path}/{fname}.tmp.gif"]).communicate()
        tf.close()   
        os.replace(f"{thumbs_path}/{fname}.tmp.gif", f"{thumbs_path}/{fname}.gif")
        
    else:
        # Thumbnail it
        out, err = Popen(["convert",tf.name,"-resize","x450","-quality","75",f"{thumbs_path}/{fname}.tmp.jpg"]).communicate()
        tf.close()   
        os.replace(f"{thumbs_path}/{fname}.tmp.jpg", f"{thumbs_path}/{fname}.jpg")
    
    return getThumbsPath(fname, full=False)
    

def buildPageDict(url, headers):
    ''' Take URL etc and create a dict using a standardised format
    
    This function is the authority on the structure of that format
    '''
    r = {
        "key" : url,
        "domain" : "",
        "path" : "",
        "filename" : "",
        "bytes" : False,
        "last-mod" : "FFFFFF",
        "content-type" : "FFFFFF",
        "title" : "",
        "valid-at" : time.strftime('%Y-%m-%d'),
        "image-path" : False,
        "keywords" : []
        }
    
    if headers.get("Last-Modified") is not None:
        r["last-mod"] = headers.get("Last-Modified")
        
    if headers.get("Content-Type") is not None:
        r["content-type"] = headers.get("content-type").lower()
        
    if headers.get("Content-Length") is not None:
        r["bytes"] = headers.get("Content-Length")
    
    r["hashkey"] = getUrlHash(url)
    
    linkparsed = urlparse(url)
    r["domain"] = linkparsed.netloc
    r["path"] = linkparsed.path
    r["filename"] = r["path"].split("/")[-1]
    
    return r

