#!/usr/bin/env python3
#
# Tombstone module
#
# Handles management of the index tombstone, including locking etc
'''
Copyright (c) 2024 B Tasker

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

    Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import os
import time

def lock(f):
    ''' Lock the tombstone
    
    Return: is_locked, locked_by_you
    '''    
    if not os.path.exists(f"{f}.tomb.lock"):
        try:
            with open(f"{f}.tomb.lock", 'w') as f:
                f.write(str(time.time()))
            return True, True
        except:
            return False, False
    else:
        # already locked
        return True, False


def unlock(f):
    ''' Unlock the tombstone
    
    As well as removing the lock, we need to check
    for the existence of a temp file and replay that
    into the file
    '''
    if not os.path.exists(f"{f}.tomb.lock"):
        # Not actually locked
        return True
    
    # Writes are now going into the tombstone
    # check for temporary writes and replay them
    temp_tomb = getTempName(f)
    
    if not temp_tomb:
        # Nothing to replay
        os.remove(f"{f}.tomb.lock")
        return True
    
    # Read the temporary tomb in then write it out to the tomb
    with open(temp_tomb, 'r') as in_f:
        for line in in_f.read().splitlines():
            print(line)
            append(f, line, ignore_lock=True)
    
    # Remove the lock
    os.remove(f"{f}.tomb.lock")
    
    # Remove the temporary file
    os.remove(temp_tomb)
    
    return True


def append(idx_path, line, ignore_lock=False):
    ''' Append a line to the tombstone
    
    idx_path should be the path to the index
    '''
    tomb_name = f"{idx_path}.tomb"
    locked, by_me = lock(idx_path)
    
    if not locked and not ignore_lock:
        print("Error: Failed to acquire lock. Forcing use of temp file")
    
    if not by_me and not ignore_lock:
        # File is already locked, write to the
        # temp tombstone instead
        tomb_name = getTempName(idx_path, False)
    
    with open(tomb_name, 'a+') as f:
        f.write(f"{line}\n")

    if by_me and not ignore_lock:
        unlock(idx_path)


def remove(idx_path):
    ''' Remove the tombstone
    '''
    if os.path.exists(f"{idx_path}.tomb"):
        os.remove(f"{idx_path}.tomb")
    

def getPaths(idx_path):
    ''' Get the path to any relevant tombstones
    '''
    paths = [f"{idx_path}.tomb"]
    
    temp_path = getTempName(idx_path)
    if temp_path:
        paths.append(temp_path)
    
    return paths


def getTempName(idx_path, check_exist = True):
    ''' Get the name of the temporary tombstone
    
    There are 2 modes:
    
    * check_exist: False - tell us what the name would be
    * check_exist: True - check if it exists and tell us the name if so
    '''
    tmpname = f"{idx_path}.tomb.tmp"
    
    if not check_exist:
        return tmpname
    
    if os.path.exists(tmpname):
        return tmpname
    
    return False
